/*
    This file is part of RShape.

    RShape is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RShape is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with RShape.  If not, see <http://www.gnu.org/licenses/>.

   Copyright (C) 2008-2010 Neil Birkbeck
*/
package com.nacb.rshape;

import java.awt.*;
import java.util.*;

import java.awt.geom.Line2D;
import java.awt.geom.RoundRectangle2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;

/**
 * A contour (part of a polygon).  E.g., a polygon consists of a number
 * of contours (essentially polygons).  Allows a polygon to have holes,
 * which is a contour with the opposite winding.
 * 
 * @author Neil Birkbeck.
 */
public class Contour{

    BezTriple [] points;

    public Contour(){ }

    /**
     * Create a contour from the string encoding.
     * The string encoding is an svg format.
     *
     * @param string  the data to extract the object from.
     */
    public Contour(String string){
	String s [] = string.split(" ");
	int npoints = new Integer(s[0]).intValue();

	for(int i=0; i<s.length; i++){
	    s[i] = s[i].trim();
	}

	points = new BezTriple[npoints];

	for(int i=0; i<npoints; i++){
	    points[i] = new BezTriple(new Point2D((new Double(s[6*i+3])).doubleValue(),
						  (new Double(s[6*i+4])).doubleValue()),
				      
				      new Point2D((new Double(s[6*i+1])).doubleValue(),
						  (new Double(s[6*i+2])).doubleValue()),
				      
				      new Point2D((new Double(s[6*i+5])).doubleValue(),
						  (new Double(s[6*i+6])).doubleValue()));	    
	}
    }

    /**
     * Create a random star with the given number of points.
     * 
     * @param n the number of points in the contour.
     */
    public void createStar(int n){
	points = new BezTriple[2*n];

	double r = Random.uniform(0, Math.PI/4.0);
	double rin = Random.uniform(0.25, 0.8);
	for(int i=0; i<2*n; i++){
	    double a = 2.0*Math.PI*((double)i)/((double)n*2.0);
	    
	    double x = Math.cos(a) * ((i%2==0)?rin:1.0);
	    double y = Math.sin(a) * ((i%2==0)?rin:1.0);
	    double xu = Math.cos(r)*x - Math.sin(r)*y;
	    double yu = Math.sin(r)*x + Math.cos(r)*y;
	    
	    points[i] = new BezTriple(new Point2D(xu, yu),
				     new Point2D(xu, yu),
				     new Point2D(xu, yu));
	}
    }

    /**
     * Create a square.
     */
    public void createSquare(){
	points = new BezTriple[4];

	float [] x = {-1, 1, 1, -1};
	float [] y = {-1, -1, 1, 1};
	double r = Random.uniform(0, Math.PI/4.0);
	for(int i=0; i<4; i++){
	    double xu = Math.cos(r)*x[i] - Math.sin(r)*y[i];
	    double yu = Math.sin(r)*x[i] + Math.cos(r)*y[i];
	    
	    points[i] = new BezTriple(new Point2D(xu, yu),
				      new Point2D(xu, yu),
				      new Point2D(xu, yu));
	}		 
	
    }

    /**
     * Create a rectangle.
     */
    public void createRectangle(){
	points = new BezTriple[4];

	double w = Random.uniform(0,1);
	double h = Random.uniform(0,1);
	
	double [] x = {-w, w, w, -w};
	double [] y = {-h, -h, h, h};
	double r = Random.uniform(0, 2.0*Math.PI);

	for(int i=0; i<4; i++){
	    double xu = Math.cos(r)*x[i] - Math.sin(r)*y[i];
	    double yu = Math.sin(r)*x[i] + Math.cos(r)*y[i];
	    
	    points[i] = new BezTriple(new Point2D(xu, yu),
				      new Point2D(xu, yu),
				      new Point2D(xu, yu));
	}
    }

    /**
     * Create a random stroke (4 points long).
     */
    public void createStroke(){
	points = new BezTriple[4];

	double w = Random.uniform(0.04,1);
	double h = Random.uniform(0.04,0.04);
	
	double [] x = {-w,w,w,-w};
	double [] y = {-h,-h,h,h};
	double r = Random.uniform(0, 2.0*Math.PI);
	for(int i=0; i<4; i++){
	    double xu = Math.cos(r)*x[i] - Math.sin(r)*y[i];
	    double yu = Math.sin(r)*x[i] + Math.cos(r)*y[i];
	    
	    points[i] = new BezTriple(new Point2D(xu, yu),
				     new Point2D(xu, yu),
				     new Point2D(xu, yu));
	}
    }

    /**
     * Fix the aspect ratio of poorly defined triangles.
     * If the aspect ratio is greater than 4, then this function
     * scales the object to be closer to uniform aspect ratio.
     */
    void fixAspect(){
	double worstAngle = 0;
	double worstAspect = 1.0;
	
	Point2D center = sum();
	center = center.scale(1.0/(double)points.length);

	// Why the hell is this part in a loop ? 
	for(int i=0; i<10; i++){
	    double minx =  10e10;
	    double maxx = -10e10;
	    double miny =  10e10;
	    double maxy = -10e10;
	    double ang = Math.PI/2.0*(double)i/10.0;

	    for(int k=0; k<points.length; k++){
		Point2D p = (points[k].pt.subtract(center)).rotate(-ang);
		minx = Math.min(minx, p.getX());
		maxx = Math.max(maxx, p.getX());

		miny = Math.min(miny, p.getY());
		maxy = Math.max(maxy, p.getY());
	    }
	    double aspect  = (maxy-miny)/(maxx-minx);
	    if(Math.max(aspect, 1.0/aspect) >
	       Math.max(worstAspect, 1.0/worstAspect)){
		worstAspect = aspect;
		worstAngle = ang;
	    }
	}

	if(Math.max(worstAspect, 1.0/worstAspect)>4){
	    double aspect = worstAspect;
	    double ang = worstAngle;
		
	    double scx = aspect;
	    double scy = 1.0;

	    scx *= Random.uniform(0.75, 1.5);
	    scy *= Random.uniform(0.75, 1.5);

	    if(aspect<1.0){
		scx = 1.0;
		scy = 1.0/aspect;
	    }
	    
	    
	    for(int k=0; k<points.length; k++){		
		Point2D p = (points[k].pt.subtract(center)).rotate(-ang);
		p = new Point2D(p.getX()*scx, p.getY()*scy);
		points[k].pt = p.rotate(ang).add(center);
		points[k].pt = p.rotate(ang).add(center);
		points[k].pt = p.rotate(ang).add(center);
	    }
	}
    }

    /**
     * Create a random shape by placing a bunch of random points, connecting them,
     * and then smoothing the result a number of times.
     *
     * @param n       the number of points in th contour.
     * @param smooth  the smoothness parameter (between 0 and 1)
     * @param convex  indicator whether the object should be force to be convex (not gauranteed)
     * @param flip    whether the shape should be flipped (e.g., it is a hole in contour).
     * @param isect   the amount of non-convexity (between 0 and 1).
     */
    void createRandomPoints(int n, double smooth, boolean convex, boolean flip, double isect){
	n = Math.max(3, n);
	points = new BezTriple[n];
	
	for(int j=0; j<n; j++){
	    double x = Random.uniform(-1, 1);
	    double y = Random.uniform(-1, 1);
	    points[j] = new BezTriple(new Point2D(x,y),
				     new Point2D(x,y),
				     new Point2D(x,y));
	}

	double t = (convex?0.9:isect);
	int top = (int)Random.uniform(2,4) + (int)((1.0-isect)*20.0);
	int numSmooth = 10;

	for(int k=0; k<numSmooth; k++){
	    Point2D pointsNew [] = new Point2D[n];

	    for(int j=0; j<n; j++){
		double tuse = t*Random.uniform(0,1);
		pointsNew[j] = points[j].pt.scale(1.0-tuse)
		    .add(points[(j+1)%n].pt.add(points[(j-1+n)%n].pt).scale(tuse));
	    }
	    for(int j=0; j<n; j++){
		points[j].pt = pointsNew[j];
		points[j].prev = pointsNew[j];
		points[j].next = pointsNew[j];
	    }
	}
	fixAspect();
    }

     /**
     * Create a random shape by subdividing.  Starting with a triangle, the longest
     * edge is split into two edges, and the new vertex is pushed a long a direction
     * that is roughly perpendicular to the old edge.  Amount of motion into the object
     * is determined by the isect parameter.
     *
     * @param n       the number of points in th contour.
     * @param smooth  the smoothness parameter (between 0 and 1)
     * @param convex  indicator whether the object should be force to be convex (not gauranteed)
     * @param flip    whether the shape should be flipped (e.g., it is a hole in contour).
     * @param isect   the amount of non-convexity (between 0 and 1).
     * @see createYardy
     * @see createRandom
     */
    void create(int n, double smooth,boolean convex,boolean flip,double isect){
	n = Math.max(3, n);
	points = new BezTriple[n];
	
	Vector<BezTriple> pcur = new Vector<BezTriple>();

	for(int i=0; i<3; i++){
	    double x = Random.uniform(-1, 1);
	    double y = Random.uniform(-1, 1);
	    pcur.addElement(new BezTriple(new Point2D(x,y),
					 new Point2D(x,y),
					 new Point2D(x,y)));
	}
	
	Point2D v1 = pcur.elementAt(1).pt.subtract(pcur.elementAt(0).pt);
	Point2D v2 = pcur.elementAt(2).pt.subtract(pcur.elementAt(0).pt);

	if(v1.getX()*v2.getY() - v1.getY()*v2.getX() < 0 && !flip){
	    BezTriple p1 = pcur.elementAt(1);
	    pcur.set(1, pcur.elementAt(2));
	    pcur.set(2, p1);
	}
	   

	for(int j=3; j<n; j++){
	    Vector<BezTriple> pnew = new Vector<BezTriple>();

	    double maxdist = 0;
	    int biggest = Math.min(pcur.size()-1, (int) Random.uniform(0, pcur.size()));
	    
	    for(int i=0; i<pcur.size(); i++){
		int i1 = (i+1)%pcur.size();
		Point2D d = pcur.elementAt(i1).pt.subtract(pcur.elementAt(i).pt);
		double len = d.distance(0.0, 0.0);
		
		if(len>maxdist && Random.uniform(0,1)<0.5){
		    maxdist = len;
		    biggest = i;
		}
	    }

	    for(int i=0; i<=biggest; i++)
		pnew.addElement(pcur.elementAt(i));
	    
	    double len = maxdist;
	    int next = (biggest+1)%pcur.size();

	    Point2D diff = pcur.elementAt(next).pt.subtract(pcur.elementAt(biggest).pt);

	    //System.out.println("diff.dist:"+diff.distance(0.0, 0.0) +" :"+j +",maxdist:"+maxdist);

	    diff.normalize();
	    diff = diff.rotate(-Math.PI/2.0+Random.uniform(-isect*Math.PI/4.0, isect*Math.PI/4.0));

	    double t = Random.uniform(0.1,0.9);
	    double isectUse = 1.0-isect;
	    double scaleDisp = Math.min(1.0, 3.0*(0.5-Math.abs(t-0.5)));

	    //if(j<)
	    //isectUse *= (j-2)/4.0;
	    //isectUse = Math.min(1.5, isectUse);

	    double x = (pcur.elementAt(biggest).pt.getX()*t + pcur.elementAt(next).pt.getX()*(1.0-t)) + diff.x*Random.uniform(-len/2*isectUse*scaleDisp, len/2*scaleDisp);
	    double y = (pcur.elementAt(biggest).pt.getY()*t + pcur.elementAt(next).pt.getY()*(1.0-t)) + diff.y*Random.uniform(-len/2*isectUse*scaleDisp, len/2*scaleDisp);
	    pnew.addElement(new BezTriple(new Point2D(x,y),
					 new Point2D(x,y),
					 new Point2D(x,y)));
	    
	    
	    
	    for(int i=biggest+1; i<pcur.size(); i++)pnew.addElement(pcur.elementAt(i));
	    

	    pcur = pnew;
	}
	for(int i=0; i<pcur.size(); i++)
	    points[i] = pcur.elementAt(i);
    }
    
    void createStarBased(int n, double smooth, boolean convex, boolean flip, double isect){
	double m = Math.tan(Math.PI/n);//*(1.0 - isect);
	double mult = flip?-1:1;
	points = new BezTriple[n];
    
	for(int i=0; i<n; i++){
	    double l = Random.uniform(0.001+isect, 1);
	    Point2D p = new Point2D(l, Random.uniform(-m*l, m*l));
	    double c = Math.cos(2.0*Math.PI*i/n*mult);
	    double s = Math.sin(2.0*Math.PI*i/n*mult);
	    
	    points[i] = new BezTriple(new Point2D(p.x*c-p.y*s, p.x*s+p.y*c),
				     new Point2D(p.x*c-p.y*s, p.x*s+p.y*c),
				     new Point2D(p.x*c-p.y*s, p.x*s+p.y*c));
	    
	    /*double l = Random.uniform(0.01,1);
	      double a = Random.uniform(-Math.PI/n, Math.PI/n)+i*2*Math.PI/n;
	      points.push_back(Point2D(l*cos(a), l*sin(a)));;
	    */
	}
    }

    /**
     * Create a random shape using Brandon Yardy's method.  There is document
     * on this; it is basically a geometric approach that is similar to
     * the randomPoints method.  There are primary and secondary points.
     * so the shape has to have an even number of points.
     * 
     * @param n  the number of points in the contour.
     */
    void createYardy(int n){
	n = n + (n%2);
	int nprim = n/2;

	points = new BezTriple[n];
	for(int i=0; i<nprim; i++){
	    double x = Random.uniform(0.05, 1.0) * ((Random.uniform(0,1)<0.5)?-1:1);
	    double maxy = Math.sqrt(1.0 - x*x);
	    double y = Random.uniform(-maxy, maxy);
	    points[2*i] = new BezTriple(new Point2D(x,y),
				       new Point2D(x,y),
				       new Point2D(x,y));
	}
	for(int i=0; i<nprim; i++){
	    Point2D p1 = points[2*i].pt;
	    Point2D p2 = points[(2*i+2)%n].pt;
	    Point2D dp = p2.subtract(p1);
	    double len = p1.distance(p2) * 0.6;
	    Point2D avg = Point2D.average(p1, p2);

	    double x = Random.uniform(-len, len);
	    double maxy = Math.sqrt(len*len - x*x) * 0.5;
	    double y = Random.uniform(maxy, maxy);

	    Point2D pt = (new Point2D(x,y)).rotate(Math.atan2(dp.getY(), dp.getX()));
	    x = pt.getX() + avg.getX();
	    y = pt.getY() + avg.getY();
	    points[(2*i+1)%n] = new BezTriple(new Point2D(x,y),
					      new Point2D(x,y),
					      new Point2D(x,y));

	}
    }

    /**
     * Try to make the shape convex by seeing if each point is to the "left"
     * of the edge defined by its previous and next adjacent points.  Points
     * that don't satisfy this are pushed to the average of these neighboring
     * points.  This is done for a few iterations.
     */
    void makeConvex(){
	for(int its=0; its<3; its++)
	    for(int i=0; i<points.length; i++){
		int i0 = (i+points.length-1)%points.length;
		int i1 = (i + 1)%points.length;
		Point2D d = points[i1].pt.subtract(points[i0].pt);
		d = d.rotate(Math.PI/2.0);
		d.normalize();
		
		Point2D di = points[i].pt.subtract(points[i0].pt);
		if(di.getX()*d.getX() + di.getY()*d.getY() > 0) {
		    points[i].pt = Point2D.average(points[i0].pt, points[i1].pt);
		}
	    }	
    }

    void scale(double s){
	for(int i=0; i<points.length; i++)
	    points[i].scale(s);
    }

    void add(Point2D p){
	for(int i=0; i<points.length; i++)
	    points[i].add(p);
    }
    
    void subtract(Point2D p){
	for(int i=0; i<points.length; i++)
	    points[i].subtract(p);
    }

    /**
     * Make the bezier control points for the contour.  The control points are determined
     * by averaging the directions to the previous and next points.  These directions
     * are perturbed for kinked corners, and the smoothness determines the length of these
     * vectors used to compute the handle points.
     *
     * @param smooth    the smoothness.  It determines how far away the bezier control points
     *                  are from their primary point.
     * @param minsmooth The probability of an corner being smooth.  Non-smooth corners
     *                  will be kinks with probability 1-minsmooth
     */
    void makeBezier(double smooth, double minsmooth){
	double straight = 0.0;

	for(int i=0; i<points.length; i++){
	    int nextPoint = (i + 1) % points.length;
	    int prevPoint = (i - 1 + points.length) % points.length;

	    Point2D dfor = points[nextPoint].pt.subtract(points[i].pt);
	    Point2D dbac = points[prevPoint].pt.subtract(points[i].pt);

	    // Backward looks wrong here (but is notused because straight == 0).
	    // Point2D forward  = points[i].pt.add(dfor.scale(1.0/3.0));;
	    // Point2D backward = points[i].pt.subtract(dbac.scale(1.0/3.0));
	    
	    double lfor = dfor.normalize();
	    double lbac = dbac.normalize();

	    Point2D  dir = dfor.subtract(dbac);
	    dir.normalize();
	    
	    double smoothDecision = (Random.uniform(0,1) <= minsmooth) ? 1.0 : 0.0;

	    // Crease was wrong, causing smooth stuff tohave crases, used to go from
	    double creaseDecision = (Random.uniform(0,1) <= (1.0-minsmooth)) ? 1.0: 0.0;
	    
	    Point2D dirFor = dir.rotate(Random.uniform(-0.7, 0.7)*creaseDecision);
	    Point2D dirBac = dir.rotate(Random.uniform(-0.7, 0.7)*creaseDecision);

	    lbac *= smooth*smoothDecision;
	    lfor *= smooth*smoothDecision;
	    
	    // Used to take lin comb of X=pt+dirFor*scale, and forward: X*(1.0-straight) + forward*straight
	    points[i].next = points[i].pt.add(dirFor.scale(lfor*0.5));	    
	    points[i].prev = points[i].pt.subtract(dirBac.scale(lbac*0.5));
	}
    }

    /**
     * Update the mins and maxs (bounding-box) using this contour.
     * These should be set appropriately by the caller.
     *
     * @param mins a running total of the min coordinate.
     * @param maxs a running total of the max coordinate.
     */
    void box(Point2D mins, Point2D maxs){
	for(int i=0; i<points.length; i++){
	    mins.setLocation(Math.min(mins.getX(),
				      points[i].pt.getX()),
			     Math.min(mins.getY(),
				      points[i].pt.getY()));
	    
	    maxs.setLocation(Math.max(maxs.getX(),
				      points[i].pt.getX()),
			     Math.max(maxs.getY(),
				      points[i].pt.getY()));

	    mins.setLocation(Math.min(mins.getX(),
				      points[i].prev.getX()),
			     Math.min(mins.getY(),
				      points[i].prev.getY()));
	    
	    maxs.setLocation(Math.max(maxs.getX(),
				      points[i].prev.getX()),
			     Math.max(maxs.getY(),
				      points[i].prev.getY()));

	    mins.setLocation(Math.min(mins.getX(),
				      points[i].next.getX()),
			     Math.min(mins.getY(),
				      points[i].next.getY()));
	    
	    maxs.setLocation(Math.max(maxs.getX(),
				      points[i].next.getX()),
			     Math.max(maxs.getY(),
				      points[i].next.getY()));
	}
    }

    /**
     * @return the sum of all the points.
     */
    Point2D sum(){
	Point2D s = new Point2D(0,0);
	for(int i=0; i<points.length; i++)
	    s = s.add(points[i].pt);
	return s;
    }

    /**
     * @return the number of points in the contour.
     */
    int size(){
	return points.length;
    }

    /**
     * @return  an svg representation of the contour.
     */
    String getData(){
	if(points.length==0)return "";
	String s = new String();

	s += "M "+points[0].pt.getX()+" "+points[0].pt.getY()+" ";

	for(int i=0;i<points.length; i++){
	    int i1 = (i+1)%points.length;

	    s += "C "+points[i].next.getX()+" "+points[i].next.getY()+" " +
		+points[i1].prev.getX()+" "+points[i1].prev.getY()+" " +
		+points[i1].pt.getX()+" "+points[i1].pt.getY()+" ";
	}
	return s+" Z";
    }

    /**
     * A debug function for drawing control points.
     */
    void drawControlPoints(Graphics2D g){	
	for(int i=0; i<points.length; i++){
	    Line2D.Double line = new Line2D.Double();
	    line.setLine(points[i].pt.getX(),
			 points[i].pt.getY(),
			 points[i].next.getX(),
			 points[i].next.getY());			 

	    g.setStroke(new BasicStroke(2));
	    g.setColor(Color.blue);
	    g.draw(line);
	    
	    line.setLine(points[i].pt.getX(),
			 points[i].pt.getY(),
			 points[i].prev.getX(),
			 points[i].prev.getY());

	    g.draw(line);

	    line.setLine(points[i].pt.getX(),
			 points[i].pt.getY(),
			 points[i].pt.getX()+1,
			 points[i].pt.getY()+1);
	    g.draw(line);

	    g.setColor(Color.red);

	    Ellipse2D e = new Ellipse2D.Double();
	    e.setFrameFromDiagonal(points[i].pt.getX()-2,
				   points[i].pt.getY()-2,
				   points[i].pt.getX()+2,
				   points[i].pt.getY()+2);
	    g.draw(e);

	    g.setStroke(new BasicStroke(1));
	    RoundRectangle2D r = new RoundRectangle2D.Double();
	    r.setFrameFromDiagonal(points[i].prev.getX()-2,
				   points[i].prev.getY()-2,
				   points[i].prev.getX()+2,
				   points[i].prev.getY()+2);
	    g.draw(r);
	    
	    r.setFrameFromDiagonal(points[i].next.getX()-2,
				   points[i].next.getY()-2,
				   points[i].next.getX()+2,
				   points[i].next.getY()+2);
	    g.draw(r);
	}
    }

    /**
     * Draw the contours points into the GeneralPath
     *
     * @param path the path to draw the contour points into.
     */
    void draw(GeneralPath path){
	if(points.length==0)return;
	
	path.moveTo((float)points[0].pt.getX(),
		    (float)points[0].pt.getY());
	for(int i=0;i<points.length; i++){
	    int i1 = (i+1)%points.length;
	    path.curveTo((float)points[i].next.getX(),
			 (float)points[i].next.getY(),
			 (float)points[i1].prev.getX(),
			 (float)points[i1].prev.getY(),
			 (float)points[i1].pt.getX(),
			 (float)points[i1].pt.getY());
	}
	path.closePath();
    }

    /**
     * Helper function to evaluate the position of the curve
     * segment.
     * 
     * @param i the curve segment index.
     * @param t the distance on the segment (between 0 and 1)
     */
    public Point2D eval(int i, double t){
	int ip = (i+1)%points.length;
	double px=0, py=0;
	
	double s0 = (1.0-t)*(1.0-t)*(1.0-t);
	double s1 = 3.0*t*(1.0-t)*(1.0-t);
	double s2 = 3.0*t*t*(1.0-t);
	double s3 = t*t*t;

	Point2D pt = new Point2D(s0*points[i].pt.getX()+
				 s1*points[i].next.getX()+
				 s2*points[ip].prev.getX()+
				 s3*points[ip].pt.getX(),
				 s0*points[i].pt.getY()+
				 s1*points[i].next.getY()+
				 s2*points[ip].prev.getY()+
				 s3*points[ip].pt.getY());
	return pt;
    }

    /**
     * @return the number of smooth corners.
     */
    public int getNumberOfSmoothCorners(){
	int cnt = 0;
	for(int i=0; i<points.length; i++){
	    if(!points[i].line){
		Point2D d1 = points[i].next.subtract(points[i].pt);
		Point2D d2 = points[i].pt.subtract(points[i].prev);
		d1.normalize();
		d2.normalize();
		cnt += (d1.dot(d2) >= (0.99)) ? 1 : 0;
	    }
	}
	return cnt;
    }

    /**
     * @return the number of corners
     */
    public int getNumberOfCorners(){
	return points.length;
    }
    
    /**
     * @return the length of the contour.
     */
    public double getLength(){
	double length = 0;
	for(int i=0; i<points.length; i++){
	    Point2D p = new Point2D(points[i].pt.x, points[i].pt.y); 
	    for(double t=1.0/32.0; t<=1.0; t+=1.0/32.0){
		Point2D pt = eval(i, t);
		length +=  pt.distance(p.getX(), p.getY());
		p = pt;
	    }
	}
	return length;
    }
    
    /**
     * @return a list of the corner angles (absolute value).
     */
    public double[] getLineAngles(){
	double vec[] = new double[points.length];

	for(int i=0; i<points.length; i++){
	    int nextPoint = (i + 1) % points.length;
	    int prevPoint = (i - 1 + points.length) % points.length;
	    Point2D vf = points[nextPoint].pt.subtract(points[i].pt);
	    Point2D vb = points[prevPoint].pt.subtract(points[i].pt);

	    vf.normalize();
	    vb.normalize();

	    double a = Math.acos(vf.dot(vb));
	    vec[i] = a;
	}
	return vec;
    }
};
