/*
    This file is part of RShape.

    RShape is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RShape is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with RShape.  If not, see <http://www.gnu.org/licenses/>.

   Copyright (C) 2008-2010 Neil Birkbeck
*/
package com.nacb.rshape;

import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;
import java.awt.image.*;

/**
 * Polygon is the core class for the random polygon data.
 * 
 * @author      Neil Birkbeck
 * @version     %I%, %G%
 * @since       1.0
 */
public class Polygon extends Shape {
    Contour [] contours;

    /**
     * Create a new, default (random), polygon.
     */
    public Polygon(){
	color = new Color(0.4f,0.4f,0.4f);
	lineColor = new Color(0.5f,0.5f,0.5f);
	fill = true;

	double array [] = {0.3,0.3,0.3,0.1};
	create(array, 5, 0.5, false, 0, 0.5, 0);
    }
    
    /**
     * Instantiate the polygon from a given string
     * representation.  Only works for polygons with
     * a single contour.
     *
     * @param string  the svg representation of the polygon.
     * @see Contour(String)
     */
    public Polygon(String s){
	// Each contour is prefixed with poly
	String[] strings = s.split("poly");	
	contours = new Contour[strings.length];

	// Add each contour (if possible)
	for (int i=0; i<strings.length; i++) {
	    try {
		strings[i] = strings[i].trim();
		// System.out.println("Creating contour:" + strings[i]);
		contours[i] = new Contour(strings[i]);	
	    }
	    catch (Exception e) {
		// System.out.println("Error creating contour:" + e);
		contours[i] = new Contour();	    
	    }
	}

	// Process any of the remaining tags
	try {
	    int colorIndex = s.indexOf("color:");
	    int lineIndex = s.indexOf("line:");
	    int lineColorIndex = s.indexOf("lineColor:");
	    
	    color = new Color(1.f, 1.f, 1.f);
	    lineColor = new Color(0.f, 0.f, 0.f);
	    
	    if (colorIndex >= 0) {
		String colorString = s.substring(colorIndex + 7, colorIndex + 7 + 6);
		int colorValue = Integer.valueOf(colorString, 16);
		color = new Color((colorValue & 0xFF),
				  (colorValue >> 8) & 0xFF,
				  (colorValue >> 16) & 0xFF);
	    }
	    
	    if (lineColorIndex >= 0) {
		String colorString = s.substring(lineColorIndex + 11, lineColorIndex + 11 + 6);
		int colorValue = Integer.valueOf(colorString, 16);
		
		lineColor = new Color((colorValue & 0xFF),
				      (colorValue >> 8) & 0xFF,
				      (colorValue >> 16) & 0xFF);
	    }
	    
	    if (lineIndex >= 0) {
		String lineString = s.substring(lineIndex + 5);		
		thick = Float.valueOf(lineString);
	    }
	}
	catch (Exception e) {
	    System.out.println("Error processing line color (or thickness):" + e);
	}
    }

    /**
     * Create a square.
     */
    public void createSquare(){
	contours = new Contour[1];
	contours[0] = new Contour();
	contours[0].createSquare();
	center();
    }

    /**
     * Create a rectangle.
     */
    public void createRectangle(){
	contours = new Contour[1];
	contours[0] = new Contour();
	contours[0].createRectangle();
	center();
    }

    /**
     * Create a random squiggle (stroke).
     */
    public void createStroke(){
	contours = new Contour[1];
	contours[0] = new Contour();
	contours[0].createStroke();
    }

    /**
     * Create a shape using the star-based method.
     */
    public void createStar(int n){
	contours = new Contour[1];
	contours[0] = new Contour();
	contours[0].createStar(n);
	center();
    }

    /**
     * Create a single contour using Yardy's method.
     * 
     * @param n the number of points (will be forced to be even).
     */
    public void createYardy(int n){
	lineColor = new Color((float)Random.uniform(0,1),
			       (float)Random.uniform(0,1),
			       (float)Random.uniform(0,1));
	color = lineColor.darker();

	contours = new Contour[1];
	contours[0] = new Contour();
	contours[0].createYardy(n);

	makeBezier(0.5, 0.5);
	center();
    }
    
    /**
     * @fixme move the creation functions
     */
    public void create(double [] probs,
		       int n, double smooth,
		       boolean convex, 
		       int ncont,
		       double minsmth,
		       double isect){
	lineColor = new Color((float)Random.uniform(0,1),
			       (float)Random.uniform(0,1),
			       (float)Random.uniform(0,1));
	color = lineColor.darker();
	
	if(ncont<=0){
	    if(Random.uniform(0,1)<0.7)
		ncont = 1;
	    else if (Random.uniform(0,1)<0.7)
		ncont = 2;
	    else if (Random.uniform(0,1)<0.7)
		ncont = 3;
	    else 
		ncont = 4;
	}
	contours = new Contour[ncont];

	for(int i=0; i<ncont; i++){
	    double rnum = Random.uniform(0,1);
	    contours[i] = new Contour();
	    if(rnum < probs[0])
		contours[i].createStarBased(i==0?n:Math.max(5,n), smooth, convex, i>=1, isect);
	    else if(rnum <probs[0]+probs[1])
		contours[i].createRandomPoints(i==0?n:Math.max(5,n), smooth, convex, i>=1, isect);
	    else if(probs.length==3 || 
		    (rnum <probs[0]+probs[1]+probs[2]))
		contours[i].create(i==0?n:Math.max(5,n), smooth, convex, i>=1, isect);
	    else
		contours[i].createYardy(i==0?n:Math.max(5,n));

	    if(i>=1){
		double s = Random.uniform(0,0.5);
		double s1 = 1.0-s;
		contours[i].scale(s);
		contours[i].add(new Point2D(Random.uniform(-s1, s1), Random.uniform(-s1, s1)));
	    }
	}
	center();
	if(convex)
	    makeConvex();
	makeBezier(smooth, minsmth);
    }

    /**
     * Scale the point by the given amount.
     *
     * @param scale  the amount to scale.
     */
    public void scale(double scale){
	for(int i=0; i<contours.length; i++)
	    contours[i].scale(scale);
    }

    /**
     * Translate the shape by the given point
     *
     * @param point   the amount to translate
     */
    public void add(Point2D point){
	for(int i=0; i<contours.length; i++)
	    contours[i].add(point);
    }

    /**
     * Translate the shape by negative point.
     *
     * @param point  the point to translate with.
     */
    public void subtract(Point2D point){
	for(int i=0; i<contours.length; i++)
	    contours[i].subtract(point);
    }

    /**
     * Center the polygon around so the center of its bounding box
     * is at the origin.  Also scales the polygon to fit
     * in a [-1, 1] x [-1, 1] square.
     *
     * @fixme Change name of the function
     */
    public void center(){
	Point2D mins = new Point2D(10e10,10e10);
	Point2D maxs = new Point2D(-10e10,-10e10);
	for(int i=0; i<contours.length; i++){
	    contours[i].box(mins, maxs);
	}
	Point2D c = new Point2D((mins.getX()+maxs.getX())*0.5,
				(mins.getY()+maxs.getY())*0.5);
	Point2D p = new Point2D(1,1);
	for(int i=0; i<contours.length; i++){
	    contours[i].subtract(c);
	}
	scale(2.0/Math.max(maxs.getX()-mins.getX(),
			   maxs.getY()-mins.getY()));
    }

    /**
     * Try to make the contours convex.  Not gauranteed to 
     * give a convex result.
     */
    public void makeConvex(){
	for(int i=0; i<contours.length; i++)
	    contours[0].makeConvex();
    }

    /**
     * Compute bezier control points for the polygon.
     * Of the corners that will be smooth (probability pctsmooth),
     * The probability of being a crease is proportional to (1.0 - probsmooth)
     *
     * @param smooth     the smoothness amount of the control points
     * @param probsmooth  the probability of a corner being smooth.
     * @see Contour#makeBezier
     */
    public void makeBezier(double smooth, double probsmooth){
	for(int i=0; i<contours.length; i++)
	    contours[i].makeBezier(smooth, probsmooth);
    }

    /**
     * Render the polygon into the graphics object.
     *
     * @param graphics  the graphics object to render into.
     * @see Shape#draw(Graphics2D)
     */
    public void draw(Graphics2D graphics){
	GeneralPath path = new GeneralPath();
	for(int i=0; i<contours.length; i++){
	    contours[i].draw(path);
	}
	AffineTransform affine = graphics.getTransform();

	if(thick>0){
	    graphics.setStroke(new BasicStroke((float)(thick)));
	    graphics.setColor(lineColor);
	    graphics.draw(path);
	}

	graphics.setColor(color);

	if(fill)
	    graphics.fill(path);

	graphics.setTransform(affine);
    }

    /**
     * Draw the control points
     * 
     * @param graphics  the graphics2D object to draw to.
     */
    public void drawControlPoints(Graphics2D graphics){
	for(int i=0; i<contours.length; i++){
	    contours[i].drawControlPoints(graphics);
	}
    }

    /**
     * Return an svg representation of the polygon 
     *
     * @return svg representation of the polygon.
     */
    public String getData(){
	String s = "";
	for(int i=0; i<contours.length; i++){
	    s += contours[i].getData();
	}
	return s;
    }    

    /**
     * Return the number of contours
     * @return the number of contours.
     */
    public int getNumberOfContours(){
	return contours.length;
    }

    /**
     * Compute the number of smooth corners in the polygon.
     * Smooth corners have bezier control points that are colinear
     * With the point.
     *
     * @return   the total number of smooth corners.
     * @see Contour#getNumberOfSmoothCorners()
     */
    public int getNumberOfSmoothCorners(){
	int cnt = 0;
	for(int i=0; i<contours.length; i++){
	    cnt += contours[i].getNumberOfSmoothCorners();
	}
	return cnt;
    }

    /**
     * Return the number of corner points in the polygon
     *
     * @return   the number of corners
     */
    public int getNumberOfCorners(){
	int cnt = 0;
	for(int i=0; i<contours.length; i++){
	    cnt += contours[i].getNumberOfCorners();
	}
	return cnt;
    }

    /**
     * Compute the total length of the perimeter of the polygon;
     * computed as the sum over all contours.
     * 
     * @return   The length of the polygon perimeter.
     */
    public double getLength(){
	double len = 0;
	for(int i=0; i<contours.length; i++){
	    len += contours[i].getLength();
	}
	return len;
    }

    /**
     * Compute the area by renering the polygon into an image and
     * counting pixels.  Assumes the polygon has been scaled 
     * appropriately.
     * 
     * @param width   The width of the image buffer to use.
     * @param height  The height of the image buffer to use.
     * @return        The area of the polygon in pixels.
     */
    public double getArea(int width, int height){
	BufferedImage img = new BufferedImage(width, height, 
					      BufferedImage.TYPE_INT_ARGB);
	Graphics2D g = (Graphics2D) img.getGraphics();
	
	draw(g);

	int cnt = 0;
	for(int y=0; y<height; y++){
	    for(int x=0; x<width; x++){
		int pix = img.getRGB(x, y);
		if((pix & 0xFF)>0){
		    cnt++;
		}
	    }
	}
	return ((double)cnt);
    }


    /**
     * LineAngleStats is a helper class to hold statistic information
     * about the corner points in a polygon.
     */
    public class LineAngleStats {
	double  min;
	double  max;
	double  avg;
	int nobtuse;
	int  nacute;
	int       n;

	LineAngleStats(){
	    min = Math.PI*2.0;
	    max = 0.0;
	    avg = 0;
	    nobtuse = 0;
	    nacute = 0;
	    n = 0;
	}

	public void append(double d){
	    avg += d;
	    min = Math.min(d, min);
	    max = Math.max(d, max);
	    if(d>Math.PI/2.0)
		nobtuse++;
	    else
		nacute++;
	    n++;
	}

	void compute(){
	    avg /= n;
	}
	
	double getTotalAngle(){
	    return avg * n;
	}
    };

    /**
     * Compute the statistics of the corners in the polygon.
     * 
     * @return  The statistsics of the contours.
     @ @see LineAngleStats.
     */
    public LineAngleStats getLineAngleStats(){
	LineAngleStats stats = new LineAngleStats();

	for(int i=0; i<contours.length; i++){
	    double arr[] = contours[i].getLineAngles();
	    for(int j=0; j<arr.length; j++){
		stats.append(arr[j]);
	    }
	}
	stats.compute();
	return stats;
    }
};
