/*
    This file is part of RShape.

    RShape is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RShape is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with RShape.  If not, see <http://www.gnu.org/licenses/>.

   Copyright (C) 2008-2010 Neil Birkbeck
*/
package com.nacb.rshape;
import java.awt.*;
import java.awt.image.*;

/**
 * The grid-pack class is used to tile a bunch of shapes.
 * It uses an image to maintain a grid of pixels that are 
 * occupied by current shapes.  When a new shape is added,
 * through packShape, the grid is searched for the first open
 * location.  When found the shape is placed, and the occupancy
 * image is updated.  
 *  
 * Shapes are allowed to have a border.
 * 
 * @author Neil Birkbeck
 * 
 */
class ShapePacker {
    int width, height;
    BufferedImage imageCanvas;
    Graphics2D imageGraphics;

    /**
     * Construct a ShapePacker class with the given width and height used
     * for the occupancy buffer.
     * @param w the width of the buffer
     * @param h the height of the buffer.
     */
    ShapePacker(int w, int h) {
	width = w;
	height = h;

	imageCanvas = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
	imageGraphics = imageCanvas.createGraphics();
    }

    /**
     * Pack the shape (after scaling it) into the occupancy grid.
     * 
     * @param shape  the shape to pack
     * @param scale  the size of the shape
     * @param space  the amount of space that should be allowed around
     *               the shape.
     * @return whether packing was successful.
     */
    boolean packShape(Shape shape, double scale, int space) {
	shape.scale(scale);

	int sz = (int)Math.round(scale*2);
	BufferedImage b = new BufferedImage(sz+space, sz+space,
					    BufferedImage.TYPE_INT_ARGB);
	Graphics2D gg = b.createGraphics();
	
	shape.draw(gg);
	shape.scale(((float)(sz+space)/((float)sz)));
	shape.draw(gg);
	
	// Try all possible packing locations
	for(int y=0; y<height-sz; y++) {
	    for(int x=0; x<width-sz; x++) {
		int sp = 16;
		boolean failed = false;
		
		while(sp>=1 && !failed){
		    // Check a sparse set of points for overlap (every sp).
		    for(int yy=0; yy<sz && !failed; yy+=sp){
			for(int xx=0; xx<sz && !failed; xx+=sp){
			    if(b.getRGB(xx,yy)!=0 &&
			       imageCanvas.getRGB(x+xx,y+yy)!=0){
				failed = true;
				break;
			    }
			}
		    }
		    sp /= 2;
		}

		if(!failed){
		    //System.out.println("Found at "+x+","+y);
		    shape.add(new Point2D(x,y));
		    shape.draw(imageGraphics);
		    
		    shape.add(new Point2D(-x,-y));
		    shape.scale(((float)(sz)/((float)(sz+space))));
		    
		    shape.add(new Point2D(x,y));
		    shape.draw(imageGraphics);
		    return true;
		}
	    }
	}
	return false;
    }    
}